#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include "main.h"

extern int iterations;
extern float threshold, alpha, epsilon;
extern float image[HEIGHT][WIDTH][DEPTH];
extern short zeroLevelSet[HEIGHT][WIDTH][DEPTH];

//verify input
bool getAndVerifyInput(int argc, char *argv[]){
	if(argc != 5){
		printf("Need four inputs: iterations, threshold, epsilon, alpha \n");
		return false;
	}
	if(sscanf (argv[1], "%i", & iterations)!=1 || iterations<1) {
		printf("Need four inputs: iterations, threshold, epsilon, alpha \n");
		return false;
	}
	if(sscanf(argv[2], "%f", &threshold)!=1 || threshold >1){
		printf("Need four inputs: iterations, threshold, epsilon, alpha \n");
		return false;
	}
	if(sscanf(argv[3], "%f", &epsilon)!=1 || epsilon >1){
		printf("Need four inputs: iterations, threshold, epsilon, alpha \n");
		return false;
	}
	if(sscanf(argv[4], "%f", &alpha)!=1 || alpha>1){
		printf("Need four inputs: iterations, threshold, epsilon, alpha \n");
		return false;
	}
	return true;
}

void displayResult(Volume<float2> * v2, list<int3> seeds, int seedRadius){
	list<int3>::iterator seed;
	for(int x = 0; x < v2->getWidth(); x++) {
		for(int y = 0; y < v2->getHeight(); y++) {
			for(int z = 0; z < v2->getDepth(); z++) {
				float2 list;
				list.x = (float)zeroLevelSet[x][y][z] / 255.0f;
				int3 n(x,y,z);
				for (seed = seeds.begin(); seed != seeds.end(); seed++){
					if(sqrt((float)((seed->x-n.x)*(seed->x-n.x)+(seed->y-n.y)*(seed->y-n.y)+(seed->z-n.z)*(seed->z-n.z))) < seedRadius){
						list.y = 1.0f;
					}
				}
				v2->set(n, list);
			}
		}
	}
	v2->show(); //v2->showMIP();
}

template <class T>
void displaySeedPointsInWindow(Volume<T> * V, Volume<float2> * v2, int id, list<int3> seeds, int seedRadius){
	list<int3>::iterator seed;
	for(int x = 0; x < V->getWidth(); x++) {
		for(int y = 0; y < V->getHeight(); y++) {
			for(int z = 0; z < V->getDepth(); z++) {
				float2 list;
				if(id == 1){ //uchar
					list.x = (float)V->get(x,y,z) / 255.0f;
				}
				else if(id == 2){ //short
					list.x = ((float)V->get(x,y,z) + 1000.0f)/ 4000.0f;
				}
				else if(id == 3){ //ushort
					list.x = (float)V->get(x,y,z) / 2000.0f;		
				}
				int3 n(x,y,z);
				for (seed = seeds.begin(); seed != seeds.end(); seed++){
					if(sqrt((float)((seed->x-n.x)*(seed->x-n.x)+(seed->y-n.y)*(seed->y-n.y)+(seed->z-n.z)*(seed->z-n.z))) < seedRadius){
						list.y = 1.0f;
					}
				}
				v2->set(n, list);
			}
		}
	}
	v2->show(); //v2->showMIP();
}

void displayUcharVolume(Volume<uchar> * V, Volume<float2> * v2, list<int3> seeds, int seedRadius){
	for(int x = 0; x < V->getWidth(); x++) {
		for(int y = 0; y < V->getHeight(); y++) {
			for(int z = 0; z < V->getDepth(); z++) {
				image[x][y][z] = (int)V->get(x,y,z) / 255.0f;
				if(image[x][y][z] > 1.0f){
					image[x][y][z] = 1.0f;
				}
			}
		}
	}
	displaySeedPointsInWindow(V, v2, 1, seeds, seedRadius);
}

void displayShortVolume(Volume<short> * V, Volume<float2> * v2, list<int3> seeds, int seedRadius){
	for(int x = 0; x < V->getWidth(); x++) {
		for(int y = 0; y < V->getHeight(); y++) {
			for(int z = 0; z < V->getDepth(); z++) {
				image[x][y][z] = ((int)V->get(x,y,z) + 1000.0f)/ 4000.0f;
				if(image[x][y][z] > 1.0f){
					image[x][y][z] = 1.0f;
				}
			}
		}
	}
	displaySeedPointsInWindow(V, v2, 2, seeds, seedRadius);
}

void displayUshortVolume(Volume<ushort> * V, Volume<float2> * v2, list<int3> seeds, int seedRadius){ 
	for(int x = 0; x < V->getWidth(); x++) {
		for(int y = 0; y < V->getHeight(); y++) {
			for(int z = 0; z < V->getDepth(); z++) {
				image[x][y][z] = (int)V->get(x,y,z) / 2000.0f;
				if(image[x][y][z] > 1.0f){
					image[x][y][z] = 1.0f;
				}
			}
		}
	}
	displaySeedPointsInWindow(V, v2, 3, seeds, seedRadius);
}