#include <gtk/gtk.h> //used by SIPL to display volumes
#include <iostream>
#include <stdio.h>
#include <list>
#include <exception>
#include <stdlib.h>
#include <string.h>
#include "main.h" //Pixel, HEIGHT, WIDTH, DEPTH defined here
#include "update.h" //levelset evolvement happens here
#include "IO.h" //read, write, display, verify input

//to calculate runtime
#include <cstdio> 
#include <ctime> 
clock_t start;
double duration;

float image[HEIGHT][WIDTH][DEPTH] = {0}; //input -> data to be segmented
float phi[HEIGHT+BORDER][WIDTH+BORDER][DEPTH+BORDER] = {0};  //level set
short init[HEIGHT+BORDER][WIDTH+BORDER][DEPTH+BORDER] = {0}; //binary mask with seed points
short label[HEIGHT+BORDER][WIDTH+BORDER][DEPTH+BORDER] = {0}; //contains info about the layers
short zeroLevelSet[HEIGHT][WIDTH][DEPTH] = {0}; //output
float threshold, alpha, epsilon;
int iterations;

list<Pixel> lz;
list<Pixel> lp1;
list<Pixel> ln1;
list<Pixel> lp2;
list<Pixel> ln2;

list<Pixel> sz;
list<Pixel> sp1;
list<Pixel> sn1;
list<Pixel> sp2;
list<Pixel> sn2;

//fills init with circular seed point
void fillSphere(list<int3> seeds, float radius){
	list<int3>::iterator seed;
	for (seed = seeds.begin(); seed != seeds.end(); seed++){
		for(int i = seed->x - radius; i<seed->x + radius; i++){
		for(int j = seed->y - radius; j<seed->y + radius; j++){
		for(int k = seed->z - radius; k<seed->z + radius; k++){
			int3 n(i,j,k);
			if(sqrt((float)((seed->x-n.x)*(seed->x-n.x)+(seed->y-n.y)*(seed->y-n.y)+(seed->z-n.z)*(seed->z-n.z))) < radius){
				init[i][j][k] = 1;
			}
		}}}
	}
}

/* returns true if any neighbour of coordinates (i,j) in either
   init (id = 1) or label (id = 2) equals res */
bool checkMaskNeighbours(int i, int j, int k, int id, short res){
	if(id == 1){ //id == 1 -> init
		if(init[i+1][j][k] == res) //right neighbour
			return true;
		else if(init[i-1][j][k] == res) //left neighbour
			return true;
		else if(init[i][j+1][k] == res) //neighbour over
			return true;
		else if(init[i][j-1][k] == res) //neighbour under
			return true;
		else if(init[i][j][k+1] == res) //neighbour in front
			return true;
		else if(init[i][j][k-1] == res) //neighbour behind
			return true;
	}
	else if(id == 2){ //id == 2 -> label
		if(label[i+1][j][k] == res)
			return true;
		else if(label[i-1][j][k] == res)
			return true;
		else if(label[i][j+1][k] == res)
			return true;
		else if(label[i][j-1][k] == res)
			return true;
		else if(label[i][j][k+1] == res)
			return true;
		else if(label[i][j][k-1] == res)
			return true;
	}
	return false;
}

//add pixels to lists according to their label
void assignLabel(Pixel p, short level){
	switch(level){
	case 1:
		lp1.push_back(p);
		label[p.x][p.y][p.z] = level;
		phi[p.x][p.y][p.z] = level;
		break;
	case 2:
		lp2.push_back(p);
		label[p.x][p.y][p.z] = level;
		phi[p.x][p.y][p.z] = level;
		break;
	case -1:
		ln1.push_back(p);
		label[p.x][p.y][p.z] = level;
		phi[p.x][p.y][p.z] = level;
		break;
	case -2:
		ln2.push_back(p);
		label[p.x][p.y][p.z] = level;
		phi[p.x][p.y][p.z] = level;	
		break;
	}
}

//initializes Ln2, Ln1, Lz, Lp1, Lp2 based on seed point(s)
void setLevels(Pixel p, short level){
	if(label[p.x+1][p.y][p.z] == 3){
		assignLabel(Pixel(p.x+1, p.y, p.z), level);
	}
	if(label[p.x][p.y+1][p.z] == 3){
		assignLabel(Pixel(p.x, p.y+1, p.z), level);
	}
	if(label[p.x-1][p.y][p.z] == 3){
		assignLabel(Pixel(p.x-1, p.y, p.z), level);
	}
	if(label[p.x][p.y-1][p.z] == 3){
		assignLabel(Pixel(p.x, p.y-1, p.z), level);
	}
	if(label[p.x][p.y][p.z+1] == 3){
		assignLabel(Pixel(p.x, p.y, p.z+1), level);
	}
	if(label[p.x][p.y][p.z-1] == 3){
		assignLabel(Pixel(p.x, p.y, p.z-1), level);
	}

	if(label[p.x+1][p.y][p.z] == -3){
		assignLabel(Pixel(p.x+1, p.y, p.z), -level);
	}
	if(label[p.x][p.y+1][p.z] == -3){
		assignLabel(Pixel(p.x, p.y+1, p.z), -level);
	}
	if(label[p.x-1][p.y][p.z] == -3){
		assignLabel(Pixel(p.x-1, p.y, p.z), -level);
	}
	if(label[p.x][p.y-1][p.z] == -3){
		assignLabel(Pixel(p.x, p.y-1, p.z), -level);
	}
	if(label[p.x][p.y][p.z+1] == -3){
		assignLabel(Pixel(p.x, p.y, p.z+1), -level);
	}
	if(label[p.x][p.y][p.z-1] == -3){
		assignLabel(Pixel(p.x, p.y, p.z-1), -level);
	}

}	

void initialization(){
	list<Pixel>::iterator it;
	for (int i = 0; i<HEIGHT+BORDER; i++){
		for (int j = 0; j<WIDTH+BORDER; j++){
			for (int k = 0; k<DEPTH+BORDER; k++){
				if(init[i][j][k] == 0){
					label[i][j][k] = 3; 
					phi[i][j][k] = 3;
				}
				else{
					label[i][j][k] = -3; 
					phi[i][j][k] = -3;
				}
			}
		}
	}
	for (int i = 1; i<HEIGHT+1; i++){
		for (int j = 1; j<WIDTH+1; j++){
			for (int k = 0; k<DEPTH+1; k++){
				if(init[i][j][k] == 1 && checkMaskNeighbours(i, j, k, 1, 0) == true){
					lz.push_back(Pixel(i,j,k));
					label[i][j][k] = 0;
					phi[i][j][k]= 0;
				}
			}
		}
	}

	for (it = lz.begin(); it != lz.end(); it++){
		setLevels(*it, 1); //add to either Lp1 or Ln1		
	}

	for (it = lp1.begin(); it != lp1.end(); it++){
		setLevels(*it, 2); //add to Lp2
	}
	for (it = ln1.begin(); it != ln1.end(); it++){
		setLevels(*it, 2); //add to Ln2
	}
}

int main(int argc, char *argv[]){
	//verify input
	if(!getAndVerifyInput(argc, argv)){
		system("pause");
		return 0;
	}
	
	//Remember to change type when changing input
	//Also remember to change WIDTH, HEIGHT, DEPTH in main.h
	Volume<uchar> * V = new Volume<uchar>("aneurism.raw", 256, 256, 256);
	list<int3> seeds;
	seeds.push_back(int3(110, 107, 162));
	//add more seedPoints if needed here
	
	float seedRadius = 5;
	Volume<float2> * v2 = new Volume<float2>(V->getSize());
	
	//Remember to chose the correct one
	//displays input, and normalizes for further processing
	displayUcharVolume(V, v2, seeds, seedRadius);
	//displayShortVolume(V, v2, seeds, seedRadius);
	//displayUshortVolume(V, v2, seeds, seedRadius);
	
	fillSphere(seeds, seedRadius); //sets seed point
	initialization();
	//calculateMu(threshold); //only needed if Chan-Vese speed function is used

	list<Pixel>::iterator itt;
	printf("starting main loop of %i iteration\n", iterations);
	start = std::clock();
	for(int i=1; i<iterations+1; i++){
		prepareUpdates();
		updateLevelSets();
		if(i%200 == 0){
			printf("\niteration %i done", i);
		}
	}
	duration = (( std::clock() - start ) / (double) CLOCKS_PER_SEC)/60;
	for(itt = lz.begin(); itt != lz.end(); itt++){
		zeroLevelSet[itt->x][itt->y][itt->z] = 255;
	}

	printf("\nmain loop finished\n");
	printf("\n time used: %f", duration);

	displayResult(v2, seeds, seedRadius); //display result

	//write and store output data
	Volume<uchar> * v3 = new Volume<uchar>(V->getSize());
	for(int x = 0; x < V->getWidth(); x++) {
	for(int y = 0; y < V->getHeight(); y++) {
	for(int z = 0; z < V->getDepth(); z++) {
	int3 n(x,y,z);	
		v3->set(x,y,z, (uchar)zeroLevelSet[x][y][z]);
	}}}

	v3->save("result.raw");
	printf("\nfile stored\n");

	system("pause");
	return 0;
}